import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.DatatypeConverter;
import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;
import java.text.SimpleDateFormat;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Date;

/**
 * @author dingweiran
 * @version 2022/6/14
 */
public class DataTypeParser {
    private static final Logger log = LoggerFactory.getLogger(DataTypeParser.class);

    public static final int MYSQL_TYPE_DECIMAL = 0;
    public static final int MYSQL_TYPE_INT8 = 1;
    public static final int MYSQL_TYPE_INT16 = 2;
    public static final int MYSQL_TYPE_INT32 = 3;
    public static final int MYSQL_TYPE_FLOAT = 4;
    public static final int MYSQL_TYPE_DOUBLE = 5;
    public static final int MYSQL_TYPE_NULL = 6;
    public static final int MYSQL_TYPE_TIMESTAMP = 7;
    public static final int MYSQL_TYPE_INT64 = 8;
    public static final int MYSQL_TYPE_INT24 = 9;
    public static final int MYSQL_TYPE_DATE = 10;
    public static final int MYSQL_TYPE_TIME = 11;
    public static final int MYSQL_TYPE_DATETIME = 12;
    public static final int MYSQL_TYPE_YEAR = 13;
    public static final int MYSQL_TYPE_DATE_NEW = 14;
    public static final int MYSQL_TYPE_VARCHAR = 15;
    public static final int MYSQL_TYPE_BIT = 16;
    public static final int MYSQL_TYPE_TIMESTAMP_NEW = 17;
    public static final int MYSQL_TYPE_DATETIME_NEW = 18;
    public static final int MYSQL_TYPE_TIME_NEW = 19;
    public static final int MYSQL_TYPE_JSON = 245;
    public static final int MYSQL_TYPE_DECIMAL_NEW = 246;
    public static final int MYSQL_TYPE_ENUM = 247;
    public static final int MYSQL_TYPE_SET = 248;
    public static final int MYSQL_TYPE_TINY_BLOB = 249;
    public static final int MYSQL_TYPE_MEDIUM_BLOB = 250;
    public static final int MYSQL_TYPE_LONG_BLOB = 251;
    public static final int MYSQL_TYPE_BLOB = 252;
    public static final int MYSQL_TYPE_VAR_STRING = 253;
    public static final int MYSQL_TYPE_STRING = 254;
    public static final int MYSQL_TYPE_GEOMETRY = 255;

    private static final SimpleDateFormat formatter = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss.SSS");
    private static final DateTimeFormatter dateFormatter = DateTimeFormatter.ofPattern("yyyy-MM-dd HH:mm:ss.SSS");

    public static String parseInteger(com.tencent.subscribe.avro.Integer data) {
        return data.getValue();
    }

    public static String parseCharacter(com.tencent.subscribe.avro.Character data) throws UnsupportedEncodingException {
        byte[] value = data.getValue().array();
        String charset = data.getCharset();
        return new String(value, charset);
    }

    public static String parseDecimal(com.tencent.subscribe.avro.Decimal data) {
        return data.getValue();
    }

    public static String parseFloat(com.tencent.subscribe.avro.Float data) {
        return Double.toString(data.getValue());
    }

    public static String parseTimestamp(com.tencent.subscribe.avro.Timestamp data) {
        long timestamp = data.getTimestamp() * 1000 + data.getMillis();
        return formatter.format(new Date(timestamp));
    }

    public static String parseDateTime(com.tencent.subscribe.avro.DateTime data, String timezone, int dataTypeNumber) {
        switch (dataTypeNumber) {
            case MYSQL_TYPE_DATETIME:
            case MYSQL_TYPE_DATETIME_NEW:
                int year = data.getYear();
                int month = data.getMonth();
                int day = data.getDay();
                int hour = data.getHour();
                int minute = data.getMinute();
                int second = data.getSecond();
                int milli = 0;
                if (data.getMillis() != null) {
                    milli = data.getMillis();
                }
                if (timezone != null) {
                    ZonedDateTime time = ZonedDateTime.of(year, month, day, hour, minute, second,
                            milli * 1000000, ZoneId.of(timezone));
                    return time.format(dateFormatter);
                }
                return String.format("%04d-%02d-%02d %02d:%02d:%02d.%03d", year, month, day, hour, minute, second, milli);
            case MYSQL_TYPE_DATE:
            case MYSQL_TYPE_DATE_NEW:
                year = data.getYear();
                month = data.getMonth();
                day = data.getDay();
                return String.format("%04d-%02d-%02d", year, month, day);
            case MYSQL_TYPE_TIME:
            case MYSQL_TYPE_TIME_NEW:
                hour = data.getHour();
                minute = data.getMinute();
                second = data.getSecond();
                milli = 0;
                if (data.getMillis() != null) {
                    milli = data.getMillis();
                }
                return String.format("%02d:%02d:%02d.%03d", hour, minute, second, milli);
            case MYSQL_TYPE_YEAR:
                return Integer.toString(data.getYear());
            default:
                log.error("unsupported time type: {}", dataTypeNumber);
                return "";
        }
    }

    public static String parseBinaryGeometry(com.tencent.subscribe.avro.BinaryGeometry data) {
        return DatatypeConverter.printHexBinary(getBytes(data.getValue()));
    }

    public static String parseTextGeometry(com.tencent.subscribe.avro.TextGeometry data) {
        return data.getValue();
    }

    public static String parseBinaryObject(com.tencent.subscribe.avro.BinaryObject data) {
        return DatatypeConverter.printHexBinary(getBytes(data.getValue()));
    }

    public static String parseTextObject(com.tencent.subscribe.avro.TextObject data) {
        return data.getValue();
    }

    public static byte[] getBytes(ByteBuffer data) {
        byte[] ret = new byte[data.remaining()];
        data.get(ret);
        return ret;
    }
}
