import com.tencent.subscribe.avro.MessageType;
import com.tencent.subscribe.avro.Record;
import org.apache.commons.cli.*;
import org.apache.flink.api.common.state.ValueState;
import org.apache.flink.api.common.state.ValueStateDescriptor;
import org.apache.flink.api.common.typeinfo.Types;
import org.apache.flink.api.java.functions.KeySelector;
import org.apache.flink.configuration.Configuration;
import org.apache.flink.streaming.api.CheckpointingMode;
import org.apache.flink.streaming.api.datastream.DataStream;
import org.apache.flink.streaming.api.datastream.KeyedStream;
import org.apache.flink.streaming.api.environment.StreamExecutionEnvironment;
import org.apache.flink.streaming.api.functions.KeyedProcessFunction;
import org.apache.flink.streaming.connectors.kafka.FlinkKafkaConsumer;
import org.apache.flink.util.Collector;
import org.apache.kafka.clients.CommonClientConfigs;
import org.apache.kafka.clients.consumer.*;
import org.apache.kafka.common.TopicPartition;
import org.apache.kafka.common.config.SaslConfigs;
import org.apache.kafka.common.header.Header;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.xml.bind.DatatypeConverter;
import java.time.Instant;
import java.time.LocalDateTime;
import java.util.*;

public class SubscribeFromKafkaFlink {
    private static final Logger log = LoggerFactory.getLogger(SubscribeFromKafkaFlink.class);

    public static void main(final String[] args) {
        Option option_brokers = Option.builder()
                .required(true)
                .desc("The brokers option")
                .longOpt("brokers")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_topic = Option.builder()
                .required(true)
                .desc("The topic option")
                .longOpt("topic")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_group = Option.builder()
                .required(true)
                .desc("The group option")
                .longOpt("group")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_user = Option.builder()
                .required(false)
                .desc("The user option")
                .longOpt("user")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_password = Option.builder()
                .required(false)
                .desc("The password option")
                .longOpt("password")
                .hasArg()
                .optionalArg(false)
                .build();
        Option option_trans = Option.builder()
                .required(false)
                .desc("The trans2sql option")
                .longOpt("trans2sql")
                .build();

        Options options = new Options()
                .addOption(option_brokers)
                .addOption(option_topic)
                .addOption(option_user)
                .addOption(option_password)
                .addOption(option_group)
                .addOption(option_trans);

        CommandLineParser parser = new DefaultParser();

        try {
            CommandLine commandLine = parser.parse(options, args);

            run(
                    commandLine.getOptionValue("brokers"),
                    commandLine.getOptionValue("topic"),
                    commandLine.getOptionValue("group"),
                    commandLine.getOptionValue("user"),
                    commandLine.getOptionValue("password"),
                    commandLine.hasOption("trans2sql")
            );

        } catch (Exception exception) {
            System.out.print("error: ");
            System.out.println(exception.getMessage());
        }
    }

    private static void run(
            String brokers,
            String topic,
            String group,
            String user,
            String password,
            boolean trans2sql) throws Exception {
        Properties props = new Properties();
        props.put(ConsumerConfig.BOOTSTRAP_SERVERS_CONFIG, brokers);
        props.put(ConsumerConfig.GROUP_ID_CONFIG, group);
        props.put(ConsumerConfig.CLIENT_ID_CONFIG, group + "_" + Instant.now().getEpochSecond());
        props.put(ConsumerConfig.ENABLE_AUTO_COMMIT_CONFIG, true);
        props.put(ConsumerConfig.AUTO_COMMIT_INTERVAL_MS_CONFIG, 5000);
        props.put(ConsumerConfig.ISOLATION_LEVEL_CONFIG, "read_committed");
        props.put(ConsumerConfig.AUTO_OFFSET_RESET_CONFIG, "earliest");
        props.put(ConsumerConfig.FETCH_MAX_BYTES_CONFIG, 20 << 20); // kafka消息上限20M
        props.put(ConsumerConfig.KEY_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.StringDeserializer");
        props.put(ConsumerConfig.VALUE_DESERIALIZER_CLASS_CONFIG, "org.apache.kafka.common.serialization.ByteArrayDeserializer");

        if (null != user && !user.equals("")) {
            props.put(CommonClientConfigs.SECURITY_PROTOCOL_CONFIG, "SASL_PLAINTEXT");
            props.put(SaslConfigs.SASL_MECHANISM, "SCRAM-SHA-512");
            props.put(SaslConfigs.SASL_JAAS_CONFIG,
                    "org.apache.kafka.common.security.scram.ScramLoginModule required " +
                            "  username=\"" + user + "\"" +
                            "  password=\"" + password + "\";");
        }

        StreamExecutionEnvironment env = StreamExecutionEnvironment.getExecutionEnvironment();
        FlinkKafkaConsumer<RecordMsgObject> consumer = new FlinkKafkaConsumer<>(topic, new RecordSchema(), props);
        consumer.setStartFromGroupOffsets();
        consumer.setCommitOffsetsOnCheckpoints(true);

        DataStream<RecordMsgObject> stream = env.addSource(consumer);
        stream.keyBy((KeySelector<RecordMsgObject, String>) recordMsgObject -> String.valueOf(recordMsgObject.partition))
                .process(new SubscribeMsgProcess(trans2sql)).setParallelism(8);


        env.enableCheckpointing(5000, CheckpointingMode.AT_LEAST_ONCE);
        env.execute("dts subscribe flink consumer");
    }

    public static class SubscribeMsgProcess extends KeyedProcessFunction<String, RecordMsgObject, Object> {

        private ValueState<Long> maxSeqState;
        private final boolean trans2sql;

        public SubscribeMsgProcess(boolean trans2sql) {
            this.trans2sql = trans2sql;
        }

        @Override
        public void open(Configuration parameters) throws Exception {
            maxSeqState = getRuntimeContext().getState(new ValueStateDescriptor<Long>("maxSeqState", Types.LONG, (long) 0));
        }

        @Override
        public void processElement(RecordMsgObject record, Context context, Collector<Object> collector) throws Exception {
            Long lastPS = maxSeqState.value();
            long ps = record.partitionSeq;
            if (ps <= lastPS) {
                log.info("Duplicated message: " + record);
                return;
            } else if (ps != lastPS + 1 && lastPS != 0) {
                throw new IllegalStateException("the partition seq is discontinuous, last: " + lastPS + ", current: " + ps);
            }
            onEntry(record.partition, record.offset, ps, record.value, trans2sql);

            maxSeqState.update(ps);
        }
    }

    private static void onEntry(int partitionNum, long offset, long partitionSeq, Record record, boolean trans2sql) throws Exception {
        log.info("partition:" + partitionNum + ", offset:" + offset + ", partitionSeq:" + partitionSeq
                + ", " + record.toString());
        if (trans2sql) {
            trans2SQL(partitionNum, offset, partitionSeq, record);
        }
    }

    private static void trans2SQL(int partitionNum, long offset, long partitionSeq, Record record) throws Exception {
        if (record.getMessageType() != MessageType.HEARTBEAT && record.getMessageType() != MessageType.CHECKPOINT) {
            log.info(
                    String.format("-->[partition: %d, offset: %d, partitionSeq: %d] [%s:%s], happenedAt: %s",
                            partitionNum,
                            offset,
                            partitionSeq,
                            record.getFileName(), record.getPosition(),
                            LocalDateTime.ofInstant(Instant.ofEpochSecond(record.getTimestamp()),
                                    TimeZone.getDefault().toZoneId())));
        }

        switch (record.getMessageType()) {
            case HEARTBEAT:
                System.out.println("heartbeat: " + record.getHeartbeatTimestamp());
                break;
            case CHECKPOINT:
                System.out.println("Checkpoint: gtid=" + record.getSyncedGtid()
                        + ", file=" + record.getFileName() + ", position=" + record.getPosition());
                break;
            case BEGIN:
                System.out.println("BEGIN");
                break;
            case COMMIT:
                System.out.println("COMMIT");
                break;
            case ROLLBACK:
                System.out.println("ROLLBACK");
                break;
            case DDL:
                System.out.println(record.getSql());
                break;
            case INSERT:
                System.out.println(transInsert(record));
                break;
            case DELETE:
                System.out.println(transDelete(record));
                break;
            case UPDATE:
                System.out.println(transUpdate(record));
                break;
            default:
                throw new IllegalStateException("unsupported message type");
        }
    }

    private static String transInsert(Record record) {
        StringBuilder sql = new StringBuilder(String.format(
                "INSERT INTO `%s`.`%s` VALUES ",
                record.getSchemaName(),
                record.getTableName()));
        List row = (List) record.getNewColumns();
        List<com.tencent.subscribe.avro.Field> columns = record.getColumns();
        String sep = "";
        sql.append('(');
        for (int i = 0; i < row.size(); i++) {
            sql.append(sep);
            sql.append(decode(columns.get(i), row.get(i)));
            sep = ", ";
        }
        sql.append(')');
        return sql.toString();
    }

    private static String transDelete(Record record) {
        StringBuilder sql = new StringBuilder(String.format(
                "DELETE FROM `%s`.`%s` WHERE ",
                record.getSchemaName(),
                record.getTableName()));
        String sep = "";
        List<com.tencent.subscribe.avro.Field> columns = record.getColumns();
        List row = (List) record.getOldColumns();
        for (int i = 0; i < row.size(); i++) {
            com.tencent.subscribe.avro.Field column = columns.get(i);
            Object value = row.get(i);
            if (null == value) {
                continue;
            }

            sql.append(sep);
            sql.append(String.format("`%s`", column.getName()));
            if (value instanceof com.tencent.subscribe.avro.EmptyObject) {
                sql.append(" IS NULL");
            } else {
                sql.append(String.format(" = %s", decode(column, value)));
            }
            sep = " AND ";
        }
        sql.append(" LIMIT 1");
        return sql.toString();
    }

    private static String transUpdate(Record record) {
        StringBuilder sql = new StringBuilder(String.format(
                "UPDATE `%s`.`%s` SET ",
                record.getSchemaName(),
                record.getTableName()));

        String sep = "";
        List<com.tencent.subscribe.avro.Field> columns = record.getColumns();
        List newRow = (List) record.getNewColumns();
        for (int i = 0; i < newRow.size(); i++) {
            com.tencent.subscribe.avro.Field column = columns.get(i);
            Object value = newRow.get(i);
            if (value == null) {
                continue;
            }

            sql.append(sep);
            sql.append(String.format("`%s` = %s", column.getName(), decode(column, value)));
            sep = ", ";
        }

        List oldRow = (List) record.getOldColumns();
        sql.append(" WHERE ");
        sep = "";
        for (int i = 0; i < oldRow.size(); i++) {
            com.tencent.subscribe.avro.Field column = columns.get(i);
            Object value = oldRow.get(i);
            if (null == value) {
                continue;
            }

            sql.append(sep);
            sql.append(String.format("`%s`", column.getName()));
            if (value instanceof com.tencent.subscribe.avro.EmptyObject) {
                sql.append(" IS NULL");
            } else {
                sql.append(String.format(" = %s", decode(column, value)));
            }
            sep = " AND ";
        }
        sql.append(" LIMIT 1");
        return sql.toString();
    }

    // 该方法不对value判空，需要提前针对不同类型的dml作判空处理
    private static String decode(com.tencent.subscribe.avro.Field field, Object value) {
        if (null == value) {
            return "DEFAULT";
        }
        Class<?> clz = value.getClass();
        if (com.tencent.subscribe.avro.Integer.class.equals(clz)) {
            return DataTypeParser.parseInteger((com.tencent.subscribe.avro.Integer) value);
        } else if (com.tencent.subscribe.avro.Character.class.equals(clz)) {
            try {
                return "_binary'" + DataTypeParser.parseCharacter((com.tencent.subscribe.avro.Character) value) + "'";
            } catch (Exception e) {
                return "x'" + DatatypeConverter.printHexBinary(((com.tencent.subscribe.avro.Character) value).getValue().array()) + "'";
            }
        } else if (com.tencent.subscribe.avro.Decimal.class.equals(clz)) {
            return DataTypeParser.parseDecimal((com.tencent.subscribe.avro.Decimal) value);
        } else if (com.tencent.subscribe.avro.Float.class.equals(clz)) {
            return DataTypeParser.parseFloat((com.tencent.subscribe.avro.Float) value);
        } else if (com.tencent.subscribe.avro.Timestamp.class.equals(clz)) {
            return "_binary'" + DataTypeParser.parseTimestamp((com.tencent.subscribe.avro.Timestamp) value) + "'";
        } else if (com.tencent.subscribe.avro.DateTime.class.equals(clz)) {
            return "_binary'" + DataTypeParser.parseDateTime(
                    (com.tencent.subscribe.avro.DateTime) value,
                    null,
                    field.getDataTypeNumber()) + "'";
        } else if (com.tencent.subscribe.avro.TimestampWithTimeZone.class.equals(clz)) {
            return "_binary'" + DataTypeParser.parseDateTime(
                    ((com.tencent.subscribe.avro.TimestampWithTimeZone) value).getValue(),
                    ((com.tencent.subscribe.avro.TimestampWithTimeZone) value).getTimezone(),
                    field.getDataTypeNumber()) + "'";
        } else if (com.tencent.subscribe.avro.BinaryGeometry.class.equals(clz)) {
            return "x'" + DataTypeParser.parseBinaryGeometry((com.tencent.subscribe.avro.BinaryGeometry) value) + "'";
        } else if (com.tencent.subscribe.avro.TextGeometry.class.equals(clz)) {
            return "_binary'" + DataTypeParser.parseTextGeometry((com.tencent.subscribe.avro.TextGeometry) value) + "'";
        } else if (com.tencent.subscribe.avro.BinaryObject.class.equals(clz)) {
            return "x'" + DataTypeParser.parseBinaryObject((com.tencent.subscribe.avro.BinaryObject) value) + "'";
        } else if (com.tencent.subscribe.avro.TextObject.class.equals(clz)) {
            if ("json".equals(((com.tencent.subscribe.avro.TextObject) value).getType())) {
                return "'" + DataTypeParser.parseTextObject((com.tencent.subscribe.avro.TextObject) value) + "'";
            } else {
                return "_binary'" + DataTypeParser.parseTextObject((com.tencent.subscribe.avro.TextObject) value) + "'";
            }
        } else if (com.tencent.subscribe.avro.EmptyObject.class.equals(clz)) {
            return "NULL";
        } else {
            throw new IllegalStateException("unsupported data type");
        }
    }
}
